#include "TANGOQueue.h"

TANGOQueue::TANGOQueue(int address, double timeout) {
	address_ = address;
	alpha_ = 0.95;
	pkmax_ = 0.05;
	r_mean_ = 0.0;
	time_head_mean_ = 0.0;
	timeout_ = 0.4;
	time_max_queue_ = timeout;
	
	ifstream file;
	string value;
	vector<string> vect;
	int lineNumber = 0;
	sending_ = false;
	
	fading_ = (double*) malloc(600000*sizeof(double));

	file.open ("data/fading_clean.txt");
	if(!file.is_open()){
		cout << "Cannot open fading file, exiting" << endl;
		exit(1);
	}
	while (file.good())
	{
		getline (file, value);
		split(vect, value, ' ');
		if(vect.size() < 2 ){
			break;
		}
		*(fading_ + lineNumber) = atof(vect.at(1).c_str());
		lineNumber++;
	}
	file.close();
}

void TANGOQueue::update()
{
	time_head_waiting_ = Scheduler::instance().clock() - time_head_arrival_;	
	time_head_mean_ = alpha_*time_head_mean_ + (1-alpha_)*time_head_waiting_;
	
	double time = Scheduler::instance().clock();
	double attenuation = getFading(fmod(time+offset_,600));
	r_ = 1*max((attenuation + 11.0),0.0); 
	
	r_mean_ = alpha_*r_mean_ + (1-alpha_)*(sending_ ? r_ : 0);
	if(r_mean_ < 1)
	{
		r_mean_ = 1.0;
	}
	sending_ = false;
}

void TANGOQueue::update_omega_max()
{
	double start;
	double t_first;
	double t_second;
	if(head_)
	{
		start = hdr_cmn::access(head_)->ts_;
		t_first = Scheduler::instance().clock() - start;
		t_second = 0.02;
		time_head_max_ = max(0.01,(timeout_ - t_first - t_second));///len_);
	}
	else
	{
		time_head_max_ = time_head_waiting_*2+0.01;
	}
}

double TANGOQueue::getValue()
{
	return -1.0*log(pkmax_)*(r_/r_mean_)*(time_head_mean_/time_head_max_);
}

double TANGOQueue::getValuePF()
{
	return r_/r_mean_;
}

Packet* TANGOQueue::enque(Packet* p) {
	latencies_[hdr_cmn::access(p)->uid_] = Scheduler::instance().clock();
	if(!head_)
	{
		time_head_arrival_ = Scheduler::instance().clock();
	}
	return PacketQueue::enque(p);
}

Packet* TANGOQueue::deque() {
	time_head_arrival_ = Scheduler::instance().clock();
	if(len_ != 0){
		latencies_.erase (hdr_cmn::access(head_)->uid_);
	}
	return PacketQueue::deque();
}

vector<Packet*> TANGOQueue::clean() {
	vector<Packet*> to_remove;
	while(len_){
		if(Scheduler::instance().clock() - latencies_[hdr_cmn::access(head_)->uid_] > time_max_queue_){
			latencies_.erase (hdr_cmn::access(head_)->uid_);
			Packet *p = deque();
			to_remove.push_back(p);
		}else{
			break;
		}
	}
	return to_remove;
}

double TANGOQueue::getFading(double in)
{
	int index = (int) (1000*in);
	return *(fading_ + index);
}

int TANGOQueue::split(vector<string>& vecteur, string chaine, char separateur)
{
	vecteur.clear();
	string::size_type stTemp = chaine.find(separateur);

	while(stTemp != string::npos)
	{
		vecteur.push_back(chaine.substr(0, stTemp));
		chaine = chaine.substr(stTemp + 1);
		stTemp = chaine.find(separateur);
	}

	vecteur.push_back(chaine);

	return vecteur.size();
}

void TANGOQueue::set_timeout(double timeout){
	time_max_queue_ = timeout;
}
