/*
 * Author:  Bastien Tauran
 * Date:    4.11.2016
 * Edited:	20.12.2016
 *
 * Scheduling policies for A-HARQ module in ns-2
 *
 * This file implements a single queue of the A-HARQ scheduling module
 */

#ifndef tango_queue_h
#define tango_queue_h

#include <iostream>
#include <fstream>
#include <math.h>
#include <vector>
#include <map>
#include "queue.h"
#include "config.h"
#include "adaptive-harq.h"

/**
*	Object representing one single queue of HARQQueue

	HARQQueue (scheduler)
	+------------------+
	|                  |
	| +--------------+ |
	| |  TANGOQueue  | |
	| +--------------+ |
	|                  |
	| +--------------+ |
	| |  TANGOQueue  | |
	| +--------------+ |
	|                  |
	| +--------------+ |
	| |  TANGOQueue  | |
	| +--------------+ |
	|                  |
	+------------------+

*/
class TANGOQueue : public PacketQueue {
	public:
		TANGOQueue(int address, double timeout);	// Constructor
		void update();								// Updates the parameters of the queue
		void update_omega_max();					// Updates the value of omega
		double getValue();							// Returns the computed value
		double getValuePF();						// Returns the computed value following PF
		Packet* enque(Packet*);						// Inserts a packet in the queue
		Packet* deque();							// Removes the head of the queue
		vector<Packet*> clean();					// Removes packet if timeout exceded and return them
		void set_timeout(double);					// Sets the value of timeout
		
		int address_;								// Address of the end node of the fid
		int offset_;								// Offset of the LMS channel
		bool sending_;								// Saying if it is the sending queue
		
	protected:
		double getFading(double);					// Returns the value of LMS channel fading
		int split(vector<string>&, string, char);	// Splits a string

		double alpha_;								// Exponential mean parameter
		double time_head_arrival_;					// Time of arrival in head of queue
		double time_head_waiting_;					// Time waiting in head
		double time_head_mean_;						// Exponential mean of time in head
		double time_head_max_;						// Max time allowed in head
		double r_;									// Throughput
		double r_mean_;								// Exponential mean of throughput
		double pkmax_;								// Probability of success
		double timeout_;							// Applicative timeout of packets
		double* fading_;							// Fading of the channel
		map<int,double> latencies_;					// Time of arrival of ach packet
		double time_max_queue_;						// Timeout for PFDrop
};

#endif
