#include "TANGOScheduler.h"

static class HARQQueueClass : public TclClass {
 public:
	HARQQueueClass() : TclClass("Queue/HARQQueue") {}
	TclObject* create(int, const char*const*) {
		return (new HARQQueue);
	}
} class_harq_queue;

static class HARQQueueRRClass : public TclClass {
 public:
	HARQQueueRRClass() : TclClass("Queue/HARQQueue/RR") {}
	TclObject* create(int, const char*const*) {
		return (new HARQQueueRR);
	}
} class_harq_queue_rr;

static class HARQQueueMLWDFClass : public TclClass {
 public:
	HARQQueueMLWDFClass() : TclClass("Queue/HARQQueue/MLWDF") {}
	TclObject* create(int, const char*const*) {
		return (new HARQQueueMLWDF);
	}
} class_harq_queue_mlwdf;

static class HARQQueuePFClass : public TclClass {
 public:
	HARQQueuePFClass() : TclClass("Queue/HARQQueue/PF") {}
	TclObject* create(int, const char*const*) {
		return (new HARQQueuePF);
	}
} class_harq_queue_pf;

static class HARQQueuePFDropClass : public TclClass {
 public:
	HARQQueuePFDropClass() : TclClass("Queue/HARQQueue/PFDrop") {}
	TclObject* create(int, const char*const*) {
		return (new HARQQueuePFDrop);
	}
} class_harq_queue_pf_drop;

static class HARQQueueFIFOClass : public TclClass {
 public:
	HARQQueueFIFOClass() : TclClass("Queue/HARQQueue/FIFO") {}
	TclObject* create(int, const char*const*) {
		return (new HARQQueueFIFO);
	}
} class_harq_queue_no_queue;

void HARQHandler::handle(Event*){
	queue_->blocked_HARQ_ = false;
	queue_->resume();
}

HARQQueue::HARQQueue() : harqh_(this) {
	bind_bool("debug_", &debug_);
	blocked_HARQ_ = false;
	
	pt_queue_ = new BaseTrace;
	dropped_ = 0;
	time_max_queue_ = 0.1;
}

void HARQQueue::reset()
{
	Queue::reset();
}

int HARQQueue::command(int argc, const char*const* argv) 
{
	Tcl& tcl = Tcl::instance();
	if (argc == 2) {
		if (strcmp(argv[1], "results") == 0) {
			results();
			return(TCL_OK);
		}
	}
	if (argc == 3) {
		if (strcmp(argv[1], "trace-queue") == 0) {
			int mode;
			const char* id = argv[2];
			Tcl_Channel ch = Tcl_GetChannel(tcl.interp(), (char*)id, &mode);
			pt_queue_->channel(ch);
			if (pt_queue_->channel() == 0) {
				tcl.resultf("trace: can't attach %s for writing", id);
				return (TCL_ERROR);
			}
			return (TCL_OK);
		}
		if (strcmp(argv[1], "set-queue-size") == 0) {
			queue_size_max_ = (atoi(argv[2]) < 10 ? 10 : atoi(argv[2]));
			if(debug_)
			{
				cout << Scheduler::instance().clock() << " : Changing queue size, new value is " << queue_size_max_ << endl;
			}
			return (TCL_OK);
		}
		if (strcmp(argv[1], "set-PFDrop-timeout") == 0) {
			time_max_queue_ = atof(argv[2]);
			if(debug_)
			{
				cout << Scheduler::instance().clock() << " : Changing PFDrop timeout value, new value is " << time_max_queue_ << endl;
			}
			return (TCL_OK);
		}
	}
	if (argc == 4) {
		if (strcmp(argv[1], "set-harq") == 0) {
			if (*argv[2] == '0' || *argv[3] == '0') {
				tcl.resultf("Cannot attach NULL Agent\n");
				return(TCL_ERROR);
			}
			harqt_ = (HARQT*)TclObject::lookup(argv[2]);
			harqr_ = (HARQR*)TclObject::lookup(argv[3]);
			if(debug_)
			{
				cout << Scheduler::instance().clock() << " : Attaching agents" << endl;
			}
			return(TCL_OK);
		}
	}
	return Connector::command(argc, argv);
}

bool HARQQueue::isNewAddress(int address)
{
	for(unsigned int i = 0; i < qs_.size(); i++)
	{
		TANGOQueue* q = qs_[i];
		if(q->address_ == address)
		{
			return false;
		}
	}
	return true;
}

void HARQQueue::enque(Packet* p)
{
	if(hdr_cmn::access(p)->ptype_ == 73){
		queueFEC_->enque(p);
		if(!blocked_HARQ_){
			resume();
		}
		return;
	}
	int address = hdr_ip::access(p)->daddr();
	if(isNewAddress(address))
	{
		TANGOQueue* q = new TANGOQueue(address,time_max_queue_);
		q->offset_ = (qs_.size()*51)%600;
		qs_.push_back(q);
	}
	for(unsigned int i = 0; i < qs_.size(); i++)
	{
		TANGOQueue* q = qs_[i];
		if(q->address_ == address)
		{
			if(q->length() > queue_size_max_){
				dropped_++;
				print_trace_queue('d', hdr_ip::access(p)->flowid(), i, 1, 1);
				drop(p);
				if(!blocked_HARQ_){
					resume();
				}
				return;
			}
			q->enque(p);
			print_trace_queue('r', hdr_ip::access(p)->flowid(), i, 0, 0);
			hdr_harq::access(p)->LMS_offset_ = q->offset_;
			if(!blocked_HARQ_){
				resume();
			}
			return;
		}
	}
	if(!blocked_HARQ_){
		resume();
	}
}

Packet* HARQQueue::deque()
{
	if(blocked_HARQ_){
		return (new PacketQueue())->deque();
	}
	if(debug_)
	{
		cout << Scheduler::instance().clock() << endl;
	}
	if(queueFEC_->length() != 0){
		Packet* p = queueFEC_->deque();
		blocked_ = 1;
		blocked_HARQ_ = true;
		Scheduler::instance().schedule(&harqh_, new Event(), hdr_cmn::access(p)->size_*8.0/50000000);
		Scheduler::instance().schedule(harqr_, p, harqt_->getDelay());
		return (new PacketQueue())->deque();
	}
	for(unsigned int i = 0; i < qs_.size(); i++)
	{
		TANGOQueue* q = qs_[i];
		q->update();
		q->update_omega_max();
	}
	TANGOQueue* q = getAddressDeque();
	if(!q)
	{
		return (new PacketQueue())->deque();
	}
	q->sending_ = true;
	Packet* p = q->deque();
	blocked_ = 1;
	blocked_HARQ_ = true;
	return p;
}

void HARQQueue::print_trace_queue(char action, int fid, int index, int nb_drop, int cause){
	sprintf(pt_queue_->buffer(), "%c "TIME_FORMAT" %d %d %d %d %d",
		action,
		pt_queue_->round(Scheduler::instance().clock()),
		fid,
		qs_[index]->length(),
		index,
		nb_drop,
		cause);
	pt_queue_->dump();
}

TANGOQueue* HARQQueue::getAddressDeque()	
{
	cout << "The function getAddressDeque() from class HARQQueue must never be called" << endl;
	return NULL;
}
	
TANGOQueue* HARQQueueRR::getAddressDeque()	
{
	if(qs_.size() == 0){
		return NULL;
	}
	
	int first_index = index_current_;
	while(qs_[index_current_]->length() == 0)
	{
		index_current_++;
		index_current_ %= qs_.size();
		if(index_current_ == first_index)
		{
			return NULL;
		}
	}
	TANGOQueue* addr = qs_[index_current_];
	print_trace_queue('s', hdr_ip::access(qs_[index_current_]->head())->flowid(), index_current_, 0, 0);
	
	if(debug_)
	{
		cout << index_current_ << endl;
	}
	
	index_current_++;
	index_current_ %= qs_.size();
	return addr;
}
	
TANGOQueue* HARQQueueMLWDF::getAddressDeque()	
{
	if(qs_.size() == 0){
		return NULL;
	}
	
	int index_max = -1;
	double max = -1.0;
	
	for(unsigned int i = 0; i < qs_.size(); i++)
	{
		TANGOQueue* q = qs_[i];
		q->update();
		q->update_omega_max();
		if(q->length() == 0){
			continue;
		}
		double score = q->getValue();
		if(debug_)
		{
			cout << "Queue " << i << ": score=" << score << ", size is " << q->length() << endl;
		}
		if(score > max)
		{
			max = score;
			index_max = i;
		}
	}
	if(index_max == -1){
		return NULL;
	}
	//cout << "TANGOQueue: index chosen: " << index_max << ", address is " << qs_[index_max]->address_ << " at " << Scheduler::instance().clock() << endl;
	//cout << "-----" << endl;
	if(debug_)
	{
		cout << "TANGOQueue: index chosen: " << index_max << endl;
		cout << "-----" << endl;
	}
	print_trace_queue('s', hdr_ip::access(qs_[index_max]->head())->flowid(), index_max, 0, 0);
	return qs_[index_max];
}
	
TANGOQueue* HARQQueuePF::getAddressDeque()	
{
	if(qs_.size() == 0){
		return NULL;
	}
	
	int index_max = -1;
	double max = -1.0;
	
	for(unsigned int i = 0; i < qs_.size(); i++)
	{
		TANGOQueue* q = qs_[i];
		q->update();
		q->update_omega_max();
		if(q->length() == 0){
			continue;
		}
		double score = q->getValuePF();
		if(debug_)
		{
			cout << "Queue " << i << ": score=" << score << ", size is " << q->length() << endl;
		}
		if(score > max)
		{
			max = score;
			index_max = i;
		}
	}
	if(index_max == -1){
		return NULL;
	}
	if(debug_)
	{
		cout << "TANGOQueue: index chosen: " << index_max << endl;
		cout << "-----" << endl;
	}
	print_trace_queue('s', hdr_ip::access(qs_[index_max]->head())->flowid(), index_max, 0, 0);
	return qs_[index_max];
}
	
TANGOQueue* HARQQueuePFDrop::getAddressDeque()	
{
	if(qs_.size() == 0){
		return NULL;
	}
	
	for(unsigned int i = 0; i < qs_.size(); i++)
	{
		vector<Packet*> to_remove = qs_[i]->clean();
		int drops = to_remove.size();
		dropped_ += drops;
		if(drops){
			print_trace_queue('d', hdr_ip::access(to_remove[0])->flowid(), i, drops, 2);
		}
		for(unsigned int j = 0; j < to_remove.size(); j++){
			drop(to_remove[j]);
		}
	}
	
	int index_max = -1;
	double max = -1.0;
	
	for(unsigned int i = 0; i < qs_.size(); i++)
	{
		TANGOQueue* q = qs_[i];
		q->update();
		q->update_omega_max();
		if(q->length() == 0){
			continue;
		}
		double score = q->getValuePF();
		if(debug_)
		{
			cout << "Queue " << i << ": score=" << score << ", size is " << q->length() << endl;
		}
		if(score > max)
		{
			max = score;
			index_max = i;
		}
	}
	if(index_max == -1){
		return NULL;
	}
	if(debug_)
	{
		cout << "TANGOQueue: index chosen: " << index_max << endl;
		cout << "-----" << endl;
	}
	print_trace_queue('s', hdr_ip::access(qs_[index_max]->head())->flowid(), index_max, 0, 0);
	return qs_[index_max];
}

TANGOQueue* HARQQueueFIFO::getAddressDeque()	
{
	if(qs_.size() == 0){
		return NULL;
	}
	if(debug_)
	{
		cout << "TANGOQueue: index chosen: 0" << endl;
		cout << "-----" << endl;
	}
	if(qs_[0]->length() == 0){
		return NULL;
	}
	return qs_[0];
}

void HARQQueueFIFO::enque(Packet* p)
{
	if(hdr_cmn::access(p)->ptype_ == 73){
		queueFEC_->enque(p);
		if(!blocked_HARQ_){
			resume();
		}
		return;
	}
	if(qs_.size() == 0)
	{
		TANGOQueue* q = new TANGOQueue(hdr_ip::access(p)->daddr(),time_max_queue_);
		q->offset_ = (qs_.size()*51)%600;
		qs_.push_back(q);
	}
	if(qs_[0]->length() > queue_size_max_){
		dropped_++;
		print_trace_queue('d', hdr_ip::access(p)->flowid(), 0, 1, 1);
		drop(p);
		if(!blocked_HARQ_){
			resume();
		}
		return;
	}
	qs_[0]->enque(p);
	print_trace_queue('r', hdr_ip::access(p)->flowid(), 0, 0, 0);
	hdr_harq::access(p)->LMS_offset_ = qs_[0]->offset_;
	if(!blocked_HARQ_){
		resume();
	}
}

void HARQQueue::results()
{
	cout << "Packets dropped by the queue: " << dropped_ << endl;
}
