/*
 * Author:  Bastien Tauran
 * Date:    4.11.2016
 *
 * Adaptive-HARQ module
 *
 * Adaptation of A-HARQ to NS-2. Static case
 */

#include "adaptive-harq-static.h"

#define CANAL_DURATION 600

static class HARQRStaticClass: public TclClass {
	public:
		HARQRStaticClass() : TclClass("Agent/HARQR/Static") {}
		TclObject* create(int, const char*const*) {
			return (new HARQRStatic);
		}
} class_harqr_static;

HARQRStatic::HARQRStatic():HARQR()
{
	double SNR_decoding = -4.8612;
	initTables();
	bind_bool("debug_", &debug_);
	bind("offset_time_", &offset_time_);
	
	IM_threshold_ = getCapa(SNR_decoding);
}

double HARQRStatic::getError(int nb_trans, double IM, int offset)
{
	double capacity = getCapacity(offset);
	double a = capacity*(nb_symbols_to_send[Es_N0_ - 7][nb_trans]);
	double n_symbols = 1680*6/n_;
	double new_IM = IM + a/n_symbols;
	
	if(debug_)
	{
		cout << Scheduler::instance().clock() << " HARQRStatic: nb_trans = " << nb_trans << endl;
		cout << Scheduler::instance().clock() << " HARQRStatic: symboles transmis h2 = " << nb_symbols_to_send[Es_N0_ - 7][nb_trans] << endl;
		cout << Scheduler::instance().clock() << " HARQRStatic: Capacity a = " << a << endl;
		cout << Scheduler::instance().clock() << " HARQRStatic: New IM = " << new_IM << ", old IM = " << IM << ", threshold = " << IM_threshold_ << endl;
	}
	if(new_IM >= IM_threshold_)
	{
		if(debug_)
		{
			cout << "decoded" << endl;
		}
		return 0; // decoded
	}
	if(debug_)
	{
		cout << "not decoded" << endl;
	}
	return new_IM; // not decoded
}

int HARQRStatic::computeBitsToSend(int nb_trans)
{	
	int Es_N0 = Es_N0_ - 7;
	return (nb_symbols_to_send[Es_N0][nb_trans]*n_);
}

double HARQRStatic::getCapacity(int offset)
{	
	double time = Scheduler::instance().clock();
	
	double off = 0.0;
	if(offset_time_ == 0)
	{
		off = offset;
	}
	else
	{
		off = offset_time_;
	}
	
	double attenuation = getFading(fmod(time+off,600));
	
	if(debug_)
	{
		cout << Scheduler::instance().clock() << " HARQRStatic: Attenuation h = " << attenuation << endl;
	}
	return getCapa(Es_N0_ + attenuation);
}

int HARQRStatic::split(vector<string>& vecteur, string chaine, char separateur)
{
	vecteur.clear();
	string::size_type stTemp = chaine.find(separateur);

	while(stTemp != string::npos)
	{
		vecteur.push_back(chaine.substr(0, stTemp));
		chaine = chaine.substr(stTemp + 1);
		stTemp = chaine.find(separateur);
	}

	vecteur.push_back(chaine);

	return vecteur.size();
}

void HARQRStatic::initTables()
{	
	ifstream file;
	string value;
	vector<string> vect;
	int lineNumber = 0;
	
	fading_ = (double*) malloc(600000*sizeof(double));
	
	file.open ("data/fading_clean.txt");
	if(!file.is_open()){
		cout << "Cannot open fading file, exiting" << endl;
		exit(1);
	}
	while (file.good())
	{
		getline (file, value);
		split(vect, value, ' ');
		if(vect.size() < 2 ){
			break;
		}
		*(fading_ + lineNumber) = atof(vect.at(1).c_str());
		lineNumber++;
	}
	file.close();
	
	capa_ = (double*) malloc(131001*sizeof(double));
	
	lineNumber = 0;
	file.open ("data/capa.txt");
	if(!file.is_open())
	{
		cout << "Cannot open capacity file, exiting" << endl;
		exit(1);
	}
	while (file.good())
	{
		getline (file, value);
		split(vect, value, ' ');
		if(vect.size() < 2 ){
			break;
		}
		*(capa_ + lineNumber) = atof(vect.at(1).c_str());
		lineNumber++;
		
	}
	file.close();
}

double HARQRStatic::getCapa(double in)
{
	int index = (int) (1000*(in+91));
	return *(capa_ + index);
}

double HARQRStatic::getFading(double in)
{
	int index = (int) (1000*in);
	return *(fading_ + index);
}
