/*
 * Author:  Bastien Tauran
 * Date:    4.11.2016
 *
 * Adaptive-HARQ module
 *
 * Adaptation of A-HARQ to NS-2. Static case
 */

#ifndef HARQ_STATIC_H
#define HARQ_STATIC_H

#include "adaptive-harq.h"
#include <fstream>
#include <vector>
#include <string>
#include <math.h>

/**
* Static version of HARQ, following R. Ali Ahmad thesis work
*/
class HARQRStatic : public HARQR {
	public:
		HARQRStatic();
	protected:
		double getError(int, double, int);			// Compute the IM to know if a packet has been decoded
		int computeBitsToSend(int);					// Compute the number of bits to send at the next retransmission
		double getCapacity(int);					// Return the capacity of the channel at given time
		int split(vector<string>&, string, char);	// Split one String and put it in a vector, return its size
		void initTables();							// Initialize the tables
		double getCapa(double in);					// Compute the capacity
		double getFading(double in);				// Compute the fading
		
		double IM_threshold_;						// IM threshold to decode a packet
		double* capa_;								// Capacity of the channel
		double* fading_;							// Fading of the channel
		double offset_time_;						// Offset pointing to the beginning of the fading file
		int debug_;									// To debug
};

#endif
