/*
 * Author:  Bastien Tauran
 * Date:    4.11.2016
 * Edited:	20.12.2016
 *
 * Adaptive-HARQ module
 *
 * Adaptation of A-HARQ to NS-2
 */

#include "adaptive-harq.h"

#define BUFFER_SIZE 		500
#define TIME_MAX_FID_BUFFER	1
#define RETRY_LIMIT 		3
#define FAVOR_LIMIT 		1

int hdr_harq::offset_ = 0;

static class HARQTClass: public TclClass {
	public:
		HARQTClass() : TclClass("Agent/HARQT") {}
		TclObject* create(int, const char*const*) {
			return (new HARQT);
		}
} class_harqt;

static class HARQRClass: public TclClass {
	public:
		HARQRClass() : TclClass("Agent/HARQR") {}
		TclObject* create(int, const char*const*) {
			return (new HARQR);
		}
} class_harqr;

////////// HARQ //////////

HARQ::HARQ(packet_t pkttype):Agent(pkttype)
{
	Es_N0_ = 7;
	delay_ = 0.01;
	n_ = 2;
}

// Probability of error for each transmission and for different SNR
double HARQ::error_prob[7][4] =
{
	{0.42,0.69,0.87,1.00}, //EsN0 = 7
	{0.46,0.74,0.90,1.00}, //EsN0 = 8
	{0.49,0.76,0.91,1.00}, //EsN0 = 9
	{0.51,0.79,0.93,1.00}, //EsN0 = 10
	{0.55,0.82,0.95,1.00}, //EsN0 = 11
	{0.58,0.84,0.96,1.00}, //EsN0 = 12
	{0.61,0.87,0.97,1.00}  //EsN0 = 13
};

// Number of symbols to send at each transmission and for different SNR
/*int HARQ::nb_symbols_to_send[7][4] =
{
	{759,366,664,1121}, //EsN0 = 7
	{737,377,651,1235}, //EsN0 = 8
	{710,331,579,1380}, //EsN0 = 9
	{682,322,583,1413}, //EsN0 = 10
	{671,305,624,1400}, //EsN0 = 11
	{657,277,607,1459}, //EsN0 = 12
	{645,283,586,1486}  //EsN0 = 13
};*/ // for 1000 Bytes packets

int HARQ::nb_symbols_to_send[7][4] =
{
	{1275,614,1115,1883}, //EsN0 = 7
	{1238,633,1093,2074}, //EsN0 = 8
	{1192,556,972,2318}, //EsN0 = 9
	{1145,540,979,2373}, //EsN0 = 10
	{1127,512,1048,2352}, //EsN0 = 11
	{1103,465,1019,2451}, //EsN0 = 12
	{1083,475,984,2496}  //EsN0 = 13
}; // for 1680 = 210*8 bits packets

/*int HARQ::nb_symbols_to_send[14][4] =
{
	{6764,3268,5924,10804}, //EsN0 = 7
	{6270,2689,4974,14454}, //EsN0 = 8
	{5924,2183,4199,14454}, //EsN0 = 9
	{5696,1831,3560,15773}, //EsN0 = 10
	{5559,1322,3027,16852}, //EsN0 = 11
	{5487,955,2572,17746}, //EsN0 = 12
	{5455,639,2172,18494}, //EsN0 = 13
	{5444,387,1805,19124}, //EsN0 = 14
	{5441,207,1452,19660}, //EsN0 = 15
	{5441,95,1110,20114}, //EsN0 = 16
	{5441,36,790,20493}, //EsN0 = 17
	{5441,10,513,20796}, //EsN0 = 18
	{5441,2,296,21021}, //EsN0 = 19
	{5441,0,149,21170}  //EsN0 = 20
};*/ // for 8920 bits packets

/*int HARQ::nb_symbols_to_send[14][1] =
{
	{26760}, //EsN0 = 7
	{26760}, //EsN0 = 8
	{26760}, //EsN0 = 9
	{26760}, //EsN0 = 10
	{26760}, //EsN0 = 11
	{26760}, //EsN0 = 12
	{25096}, //EsN0 = 13
	{20677}, //EsN0 = 14
	{17169}, //EsN0 = 15
	{14385}, //EsN0 = 16
	{12181}, //EsN0 = 17
	{10442}, //EsN0 = 18
	{9076}, //EsN0 = 19
	{8014}  //EsN0 = 20
};*/ // for 8920 bits packets and 1 trans

////////// HARQT //////////

HARQT::HARQT():HARQ(PT_FEC)
{
	bind_bool("debug_", &debug_);
}

void HARQT::recv(Packet* p, Handler* h)
{
	int type;
	int ret;
	
	type = hdr_cmn::access(p)->ptype_;
	
	if(type <= 4 || type == 28 || type == 29)
	{
		if(debug_)
		{
			cout << Scheduler::instance().clock() << " HARQT: PT_TCP or UDP or CBR received, id is " << hdr_cmn::access(p)->uid_ << endl;
		}
		hdr_harq::access(p)->size_ = hdr_cmn::access(p)->size_;
		hdr_cmn::access(p)->size_ += n_*nb_symbols_to_send[Es_N0_ - 7][0]/8;
		hdr_harq::access(p)->real_size_ = hdr_cmn::access(p)->size_;
		hdr_harq::access(p)->accumulated_IM_ = 0;
		send(p,&(((HARQQueue*)queue_)->harqh_));
	}
	else if(hdr_cmn::access(p)->ptype_ == 74)
	{
		if(debug_)
		{
			cout << Scheduler::instance().clock() << " HARQT: " << hdr_cmn::access(p)->ptype_ << " received (PT_ACK_FEC=74)" << endl;
		}
		ret = hdr_harq::access(p)->nb_bits_/8;
		if(ret==0)
		{
			if(debug_)
			{
				cout << Scheduler::instance().clock() << " HARQT: received positive ACK" << endl;
			}
		}
		else if(ret==-1)
		{
			if(debug_)
			{
				cout << Scheduler::instance().clock() << " HARQT: received negative ACK" << endl;
			}
		}
		else
		{
			if(debug_)
			{
				cout << Scheduler::instance().clock() << " HARQT: FEC needed: " << ret << " for id " << hdr_harq::access(p)->for_id_ << endl;
			}
			sendFEC(hdr_harq::access(p)->for_id_,ret, p, hdr_harq::access(p)->accumulated_IM_,h);
		}
		Packet::free(p);
	}
}

void HARQT::setHARQR(HARQR* harqr)
{
	harqr_ = harqr;
	if(debug_)
	{
		cout << Scheduler::instance().clock() << " : Attaching harqr to sender" << endl;
	}
}

void HARQT::sendFEC(int id, int k, Packet* pkt, double IM, Handler* h)
{
	if(debug_)
	{
		cout << Scheduler::instance().clock() << " Send " << k << " bits FEC" << endl;
	}
	Packet* p = allocpkt();
	hdr_cmn::access(p)->ptype_ = PT_FEC;
	hdr_harq::access(p)->for_id_ = id;
	hdr_harq::access(p)->accumulated_IM_ = IM;
	hdr_ip::access(p)->saddr() = addr();
	hdr_ip::access(p)->sport() = port();
	hdr_ip::access(p)->daddr() = harqr_->addr();
	hdr_ip::access(p)->dport() = harqr_->port();
	
	hdr_cmn::access(p)->size_ = k;
	hdr_harq::access(p)->real_size_ = hdr_cmn::access(p)->size_;
	hdr_ip::access(p)->fid_ = hdr_ip::access(pkt)->fid_;
	
	queue_->enque(p);
}

void HARQT::setDelay(double del)
{
	delay_ = del;
	if(debug_)
	{
		cout << Scheduler::instance().clock() << " : Changing delay value" << endl;
	}
}

double HARQT::getDelay()
{
	return delay_;
}

void HARQT::setEsN0(int EsN0)
{
	Es_N0_ = EsN0;
	if(debug_)
	{
		cout << Scheduler::instance().clock() << " : Changing EsN0 value" << endl;
	}
}

int HARQT::command(int argc, const char*const* argv) 
{	
	Tcl& tcl = Tcl::instance();
	if (argc == 3) {
		if (strcmp(argv[1], "set-queue") == 0) {
			if (*argv[2] == '0') {
				tcl.resultf("Cannot attach NULL queue\n");
				return(TCL_ERROR);
			}
			queue_ = (Queue*)TclObject::lookup(argv[2]);
			if(debug_)
			{
				cout << Scheduler::instance().clock() << " : Attaching queue to sender" << endl;
			}
			return(TCL_OK);
		}
	}
	return Connector::command(argc, argv);
}

////////// HARQR //////////

HARQR::HARQR():HARQ(PT_ACK)
{
	int i;
	
	srand (static_cast <unsigned> (time(0)));
	bind_bool("debug_", &debug_);
	bind_bool("trace_reconstruction_", &trace_reconstruction_);
	bind_bool("trace_reordering_", &trace_reordering_);
	bind_bool("reordering_", &reordering_);
	bind_bool("favor_", &favor_);
    bind("buffer_reordering_",&buffer_reordering_filling_);
    bind("timeout_",&timeout_);
	
    buffer_reordering_filling_ = 0;
	buffer_reordering_size_ = 125;
	buffer_fid_size_ = 1250;
	dropped_ = 0;
	dropped_reordering_buffer_ = 0;
	pt_reconstruction_ = new BaseTrace;
	pt_reordering_ = new BaseTrace;
	
	forwarded_ = new int [RETRY_LIMIT + 1];

	buffer_reconstruction_ = new pkt_recv[BUFFER_SIZE];
	for(i=0;i<BUFFER_SIZE;i++)
	{
		(buffer_reconstruction_+i)->valid = false;
	}
	
	buffer_reordering_ = new pkt_recv[buffer_reordering_size_];
	for(i=0;i<buffer_reordering_size_;i++)
	{
		(buffer_reordering_+i)->valid = false;
	}
	
	expected_list_ = new expected_struct[buffer_fid_size_];
	for(i=0;i<buffer_fid_size_;i++)
	{
		(expected_list_+i)->valid = false;
	}
	
	for(i=0;i<=RETRY_LIMIT;i++)
	{
		forwarded_[i] = 0;
	}
}

int HARQR::command(int argc, const char*const* argv) 
{
	int i;		
	
	Tcl& tcl = Tcl::instance();
	if (argc == 3) {
		if (strcmp(argv[1], "set-sender") == 0) {
			if (*argv[2] == '0') {
				tcl.resultf("Cannot attach NULL HARQTAgent\n");
				return(TCL_ERROR);
			}
			harqt_ = (HARQT*)TclObject::lookup(argv[2]);
			harqt_->setHARQR(this);
			if(debug_)
			{
				cout << Scheduler::instance().clock() << " : Attaching harqt to receiver" << endl;
			}
			return(TCL_OK);
		}
		if (strcmp(argv[1], "set-EsN0") == 0) {
			Es_N0_ = atoi(argv[2]);
			harqt_->setEsN0(Es_N0_);
			if(debug_)
			{
				cout << Scheduler::instance().clock() << " : Changing EsN0 value, new value is " << Es_N0_ << endl;
			}
			return(TCL_OK);
		}
		if (strcmp(argv[1], "set-delay") == 0) {
			delay_ = atof(argv[2]);
			harqt_->setDelay(delay_);
			if(debug_)
			{
				cout << Scheduler::instance().clock() << " : Changing delay value, new value is " << delay_ << endl;
			}
			return(TCL_OK);
		}
		if (strcmp(argv[1], "set-buffer-reordering-size") == 0) {
			buffer_reordering_size_ = atoi(argv[2]);
			buffer_reordering_ = new pkt_recv[buffer_reordering_size_];
			for(i=0;i<buffer_reordering_size_;i++)
			{
				(buffer_reordering_+i)->valid = false;
			}
			if(debug_)
			{
				cout << Scheduler::instance().clock() << " : Changing buffer out size, new value is " << buffer_reordering_size_ << endl;
			}
			return(TCL_OK);
		}
		if (strcmp(argv[1], "set-buffer-fid-size") == 0) {
			buffer_fid_size_ = atoi(argv[2]);
			expected_list_ = new expected_struct[buffer_fid_size_];
			for(i=0;i<buffer_fid_size_;i++)
			{
				(expected_list_+i)->valid = false;
			}
			if(debug_)
			{
				cout << Scheduler::instance().clock() << " : Changing buffer fid size, new value is " << buffer_fid_size_ << endl;
			}
			return(TCL_OK);
		}
		if (strcmp(argv[1], "trace-reconstruction") == 0) {
			int mode;
			const char* id = argv[2];
			Tcl_Channel ch = Tcl_GetChannel(tcl.interp(), (char*)id, &mode);
			pt_reconstruction_->channel(ch);
			if (pt_reconstruction_->channel() == 0) {
				tcl.resultf("trace: can't attach %s for writing", id);
				return (TCL_ERROR);
			}
			return (TCL_OK);
		}
		if (strcmp(argv[1], "trace-reordering") == 0) {
			int mode;
			const char* id = argv[2];
			Tcl_Channel ch2 = Tcl_GetChannel(tcl.interp(), (char*)id, &mode);
			pt_reordering_->channel(ch2);
			if (pt_reordering_->channel() == 0) {
				tcl.resultf("trace: can't attach %s for writing", id);
				return (TCL_ERROR);
			}
			return (TCL_OK);
		}
	}
	if (argc == 2) {
		if (strcmp(argv[1], "results") == 0) {
			results();
			return(TCL_OK);
		}
	}
	return Connector::command(argc, argv);
}

void HARQR::recv(Packet* p, Handler* h)
{
	int type;
	int index;
	double decoded;
	int nb_bits;
	int ack_sent;
	Packet* pkt;
	
	type = hdr_cmn::access(p)->ptype_;
	
	if(type <= 2 || type == 28 || type == 29)
	{
		if(debug_)
		{
			cout << Scheduler::instance().clock() << " HARQR: PT_TCP or UDP or CBR received, id is " << hdr_cmn::access(p)->uid_ << endl;
		}
		index = addToBuffer(p, h, getFirstFree());
		decoded = getError(0, hdr_harq::access(p)->accumulated_IM_, hdr_harq::access(p)->LMS_offset_); //0 = decoded
		if(trace_reconstruction_)
		{
			print_trace_reconstruction(p, type, 'r', decoded == 0 ? 1 : 0, 0, 0);
		}
		hdr_cmn::access(p)->size_ = hdr_harq::access(p)->size_;
		if(decoded == 0)
		{
			if(trace_reconstruction_)
			{
				print_trace_reconstruction(p, type, 's', 1, 0, 0);
			}
			forward_pkt(p,h,hdr_tcp::access(p)->seqno(),hdr_ip::access(p)->flowid());
			removeFromBuffer(index);
			if(debug_)
			{
				cout << Scheduler::instance().clock() << " HARQR: packet fully decoded" << endl;
			}
			ack(p,hdr_cmn::access(p)->uid_,0,0);
			forwarded_[0]++;
		}
		else
		{
			if(0==RETRY_LIMIT)
			{
				if(trace_reconstruction_)
				{
					print_trace_reconstruction(p, type, 'd', 0, 0, 0);
				}
				forward_pkt(0,0,hdr_tcp::access(p)->seqno(),hdr_ip::access(p)->flowid());
				ack(p,hdr_cmn::access(p)->uid_,-1,0);
				Packet::free(p);
				removeFromBuffer(index);
				if(debug_)
				{
					cout << Scheduler::instance().clock() << " HARQR: packet not decoded, id is " << hdr_harq::access(p)->for_id_ << endl;
				}
				dropped_++;
			}
			else
			{
				if(debug_)
				{
					cout << Scheduler::instance().clock() << " HARQR: asking for more FEC, id is " << hdr_cmn::access(p)->uid_ << endl;
				}
				nb_bits = computeBitsToSend(1);
				(buffer_reconstruction_+index)->number_ack_sent = 1;
				ack(p,hdr_cmn::access(p)->uid_,nb_bits,decoded);
			}
		}
	}
	else if(type == 73)
	{
		if(debug_)
		{
			cout << Scheduler::instance().clock() << " HARQR: " << type << " received (73=FEC), for id=" << hdr_harq::access(p)->for_id_ << endl;
		}
		index = already(p);
		if(index==-1)
		{
			cout << "Error, FEC id should already be in HARQR buffer" << endl;
		}
		ack_sent = (buffer_reconstruction_+index)->number_ack_sent;
		pkt = (buffer_reconstruction_+index)->p;
		decoded = getError(ack_sent, hdr_harq::access(p)->accumulated_IM_, hdr_harq::access(p)->LMS_offset_);//0 = decoded
		if(trace_reconstruction_)
		{
			print_trace_reconstruction(pkt, type, 'r', decoded == 0 ? 1 : 0, ack_sent, hdr_cmn::access(p)->size());
		}
		if(decoded == 0)
		{
			if(trace_reconstruction_)
			{
				print_trace_reconstruction(pkt, hdr_cmn::access(pkt)->ptype_, 's', 1, ack_sent, 0);
			}
			forward_pkt(pkt,0,hdr_tcp::access(pkt)->seqno(),hdr_ip::access(pkt)->flowid());
			if(debug_)
			{
				cout << Scheduler::instance().clock() << " HARQR: packet fully decoded, number of retransmissions: " << ack_sent << endl;
			}
			forwarded_[(buffer_reconstruction_+index)->number_ack_sent]++;
			removeFromBuffer(index);
			ack(p,hdr_harq::access(p)->for_id_,0,0);
		}
		else
		{
			if(ack_sent==RETRY_LIMIT)
			{
				if(trace_reconstruction_)
				{
					print_trace_reconstruction(pkt, hdr_cmn::access(pkt)->ptype_, 'd', 0, ack_sent, 0);
				}
				forward_pkt(0,0,hdr_tcp::access(pkt)->seqno(),hdr_ip::access(pkt)->flowid());
				
				Packet::free(pkt);
				removeFromBuffer(index);
				if(debug_)
				{
					cout << Scheduler::instance().clock() << " HARQR: packet not decoded, id is " << hdr_harq::access(p)->for_id_ << endl;
				}
				ack(p,hdr_harq::access(p)->for_id_,-1,0);
				dropped_++;
			}
			else
			{
				if(debug_)
				{
					cout << Scheduler::instance().clock() << " HARQR: asking for more FEC, id is " << hdr_harq::access(p)->for_id_ << endl;
				}
				nb_bits = computeBitsToSend(ack_sent + 1);
				((buffer_reconstruction_+index)->number_ack_sent)++;
				ack(p,hdr_harq::access(p)->for_id_,nb_bits,decoded);
			}
		}
		Packet::free(p);
	}
}

void HARQR::ack(Packet* opkt, int id, int number, double IM)
{
	Packet* npkt = allocpkt();
	
	// Header common
	hdr_cmn* ncmn = hdr_cmn::access(npkt);
	
	ncmn->ptype_ = PT_FEC_ACK;

	// Header TCP
	hdr_tcp *otcp = hdr_tcp::access(opkt);
	hdr_tcp *ntcp = hdr_tcp::access(npkt);

	ntcp->ts_echo() = otcp->ts();
	ntcp->seqno() = otcp->seqno();
	ntcp->ts() = Scheduler::instance().clock();

	// Header IP
	hdr_ip* oip = hdr_ip::access(opkt);
	hdr_ip* nip = hdr_ip::access(npkt);
	
	nip->flowid() = oip->flowid();

	// Header flags
	hdr_harq* nharq = hdr_harq::access(npkt);

	// Header HARQ
	nharq->nb_bits_ = number;
	nharq->for_id_ = id;
	nharq->accumulated_IM_ = IM;
	
	ncmn->size_ = 40;
	
	Scheduler::instance().schedule(harqt_, npkt, delay_);
}

int HARQR::getFirstFree()
{
	int i;
	
	for(i=0;i<BUFFER_SIZE;i++)
	{
		if((buffer_reconstruction_+i)->valid == false)
		{
			return i;
		}
	}
	return -1;
}

int HARQR::already(Packet* p)
{
	int i;
	int id;
	
	id = hdr_harq::access(p)->for_id_;
	for(i=0;i<BUFFER_SIZE;i++)
	{
		if((buffer_reconstruction_+i)->valid == true)
		{
			if(hdr_cmn::access((buffer_reconstruction_+i)->p)->uid_ == id)
			{
				return i;
			}
		}
	}
	return -1;
}

int HARQR::addToBuffer(Packet* p, Handler* h, int index)
{
	(buffer_reconstruction_+index)->valid = true;
	(buffer_reconstruction_+index)->p = p;
	(buffer_reconstruction_+index)->h = h;
	(buffer_reconstruction_+index)->id = hdr_cmn::access(p)->uid_;
	(buffer_reconstruction_+index)->number_ack_sent = 0;
	
	return index;
}

void HARQR::removeFromBuffer(int index)
{
	(buffer_reconstruction_+index)->valid = false;
	(buffer_reconstruction_+index)->p = NULL;
	(buffer_reconstruction_+index)->h = NULL;
}

void HARQR::forward_pkt(Packet* p,Handler* h, int seqno, int fid)
{
	if(!reordering_)
	{
		if (p != 0)
		{
			send(p,h);
			if(debug_)
			{
				cout << Scheduler::instance().clock() << " HARQR: packet forwarded " << endl;
			}
		}
		return;
	}

	int i,j;
	int index_fid;
	bool found = false;
	bool full = true;
	bool res;
	int index_fid_free = -1;
	expected_struct exp;
	pkt_recv elt = {0,0,0,0,0,0,0};
	
	if(p == 0){
		for(i=0;i<buffer_reordering_size_;i++)
		{
			if((buffer_reordering_+i)->valid == true)
			{
				for(j=0;j<buffer_fid_size_;j++)
				{
					exp = *(expected_list_+j);
					if(exp.valid==true && exp.id==(buffer_reordering_+i)->id)
					{
						if((buffer_reordering_+i)->seqno >= (expected_list_+j)->expected){
							(expected_list_+j)->expected = (buffer_reordering_+i)->seqno + 1;
							break;
						}
					}
				}
				if(debug_)
				{
					cout << Scheduler::instance().clock() << " HARQR: flushing packet " << (buffer_reordering_+i)->p << ", with fid " << (buffer_reordering_+i)->id << endl;
				}
				send((buffer_reordering_+i)->p,(buffer_reordering_+i)->h);
				buffer_reordering_filling_--;
				(buffer_reordering_+i)->valid = false;
			}
		}
		return;
	}
	
	for(i=0;i<buffer_fid_size_;i++)
	{
		exp = *(expected_list_+i);
		if(exp.valid==true)
		{
			if(exp.id==fid)
			{
				(expected_list_+i)->last_time = Scheduler::instance().clock();
				index_fid = i;
				found  = true;
				break;
			}
		}
		else
		{
			if(index_fid_free == -1)
			{
				index_fid_free = i;
			}
			full = false;
		}
	}
	if(!found)
	{
		if(full)
		{
			if(debug_)
			{
				cout << Scheduler::instance().clock() << " HARQR: ERROR FID BUFFER OUT FULL" << endl;
			}
			if(trace_reordering_)
			{
				print_trace_reordering(p,'d',0,1);
			}
			drop(p);
			dropped_reordering_buffer_++;
			return;
		}
		else
		{
			expected_struct expec;
			expec.id = fid;
			expec.expected = seqno;
			expec.last_time = Scheduler::instance().clock();
			expec.valid = true;
		
			*(expected_list_+index_fid_free) = expec;
			index_fid = index_fid_free;
		}
	}
	if(debug_)
	{
		cout << Scheduler::instance().clock() << " HARQR: packet to forward " << seqno << ", expected is " << (expected_list_+index_fid)->expected << endl;
	}
	if(trace_reordering_)
	{
		print_trace_reordering(p,'r',isFavor(fid,1),0);
	}
	if(seqno == (expected_list_+index_fid)->expected)
	{
		res = true;
		do
		{
			if(trace_reordering_)
			{
				print_trace_reordering(p,'s',isFavor(fid,1),0);
			}
			send(p,h);
			if(debug_)
			{
				cout << Scheduler::instance().clock() << " HARQR: packet forwarded " << (expected_list_+index_fid)->expected << endl;
			}
			pkt_recv ret;
			while(deque(&ret,(expected_list_+index_fid)->expected,fid))
			{
				if(trace_reordering_)
				{
					print_trace_reordering(ret.p,'d',0,1);
				}
				drop(ret.p);
				dropped_reordering_buffer_++;
			}
			((expected_list_+index_fid)->expected)++;
			res = deque(&ret,(expected_list_+index_fid)->expected,fid);
			if(res)
			{
				p = ret.p;
				h = ret.h;
			}
		}
		while(res);
	}
	else
	{
		if(debug_)
		{
			cout << Scheduler::instance().clock() << " HARQR: incorrect seqno expected " << (expected_list_+index_fid)->expected << ", but got " << seqno << endl;
		}
		if(seqno>(expected_list_+index_fid)->expected)
		{
			elt.valid = true;
			elt.p = p;
			elt.h = h;
			elt.id = fid;
			elt.seqno = seqno;
			elt.time_in = Scheduler::instance().clock();
			if(isFull()==-1)
			{
				send(p,h);
				(expected_list_+index_fid)->expected = seqno + 1;
				for(i=0;i<buffer_reordering_size_;i++)
				{
					if((buffer_reordering_+i)->valid == true)
					{
						for(j=0;j<buffer_fid_size_;j++)
						{
							exp = *(expected_list_+j);
							if(exp.valid==true && exp.id==(buffer_reordering_+i)->id)
							{
								if((buffer_reordering_+i)->seqno >= (expected_list_+j)->expected){
									(expected_list_+j)->expected = (buffer_reordering_+i)->seqno + 1;
									break;
								}
							}
						}
						if(debug_)
						{
							cout << Scheduler::instance().clock() << " HARQR: flushing packet " << (buffer_reordering_+i)->p << ", with fid " << (buffer_reordering_+i)->id << endl;
						}
						send((buffer_reordering_+i)->p,(buffer_reordering_+i)->h);
						buffer_reordering_filling_--;
						(buffer_reordering_+i)->valid = false;
					}
				}
				return;
			}
			else
			{
				if(enque(elt))
				{
					(expected_list_+index_fid)->last_time = Scheduler::instance().clock();
				}
			}
		}
		else
		{
			if(debug_)
			{
				cout << Scheduler::instance().clock() << " HARQR: seqno (" << seqno << ") < expected (" << (expected_list_+index_fid)->expected << ")" << endl;
			}
			if(trace_reordering_)
			{
				print_trace_reordering(p,'s',isFavor(fid,1),0);
			}
			send(p,h);
		}
	}
	if(debug_)
	{
		cout << Scheduler::instance().clock() << " HARQR: cleaning packets" << endl;
	}
	for(i=0;i<buffer_reordering_size_;i++)
	{
		if((buffer_reordering_+i)->valid == true)
		{
			if(Scheduler::instance().clock() > (buffer_reordering_+i)->time_in + timeout_)
			{
				buffer_reordering_filling_--;
				(buffer_reordering_+i)->valid = false;
				if(trace_reordering_)
				{
					print_trace_reordering((buffer_reordering_+i)->p,'d',isFavor((buffer_reordering_+i)->id,1),2);
				}
				drop((buffer_reordering_+i)->p);
				dropped_reordering_buffer_++;
				if(debug_)
				{
					cout << Scheduler::instance().clock() << " HARQR: packet timeout " << (buffer_reordering_+i)->seqno << ", fid is " << (buffer_reordering_+i)->id << ". Time in buffer is " << Scheduler::instance().clock() - (buffer_reordering_+i)->time_in << "s" << endl;
				}
			}
		}
	}
	if(debug_)
	{
		cout << Scheduler::instance().clock() << " HARQR: cleaning fid" << endl;
	}
	for(i=0;i<buffer_fid_size_;i++)
	{
		if((expected_list_+i)->valid == true)
		{
			if(Scheduler::instance().clock() > (expected_list_+i)->last_time + TIME_MAX_FID_BUFFER)
			{
				(expected_list_+i)->valid = false;
				int id = (expected_list_+i)->id;
				pkt_recv pkt;
				deque(&pkt, -1, id);
				if(debug_)
				{
					cout << Scheduler::instance().clock() << " HARQR: fid timeout " << (expected_list_+i)->id << ", last time is " << (expected_list_+i)->last_time << endl;
				}
			}
		}
	}
	if(debug_)
	{
		cout << Scheduler::instance().clock() << " HARQR: end cleaning" << endl;
	}
}

bool HARQR::isFavor(int fid, int offset)
{
	int i;
	int k = 0;
	int size = 0;
	pkt_recv pkt = *(new pkt_recv);
	
	for(i=0;i<buffer_reordering_size_;i++)
	{
		pkt = *(buffer_reordering_ + i);
		if(pkt.valid == true)
		{
			if(pkt.id == fid)
			{
				k++;
			}
			size++;
		}
	}
	if(debug_)
	{
		cout << Scheduler::instance().clock() << " HARQR: number of packets in queue for fid " << fid << ": " << k << endl;
		cout << Scheduler::instance().clock() << " HARQR: fid buffer size: " << size << endl;
	}
	if(k + offset <= FAVOR_LIMIT)
	{
		return true;
	}
	return false;
}

int HARQR::isFull()
{
	int i;
	for(i=0;i<buffer_reordering_size_;i++)
	{
		if((buffer_reordering_+i)->valid == false)
		{
			return i;
		}
	}
	return -1;
}

bool HARQR::enque(pkt_recv elt)
{
	int i;
	int full;
	bool favor;
	
	if(favor_)
	{
		favor = isFavor(elt.id,1);
		if(favor)
		{
			full = isFull();
			if(full == -1)
			{
				cout << "Error: no enque if buffer full" << endl;
				exit(0);
			}
			else
			{
				*(buffer_reordering_+full) = elt;
				buffer_reordering_filling_++;	
				if(debug_)
				{
					cout << Scheduler::instance().clock() << Scheduler::instance().clock() << " HARQR: packet put in buffer " << elt.seqno << endl;
				}
				return true;
			}
		}
		else
		{
			full = isFull();
			if(full == -1)
			{
				if(trace_reordering_)
				{
					print_trace_reordering(elt.p,'d',0,1);
				}
				drop(elt.p);
				dropped_reordering_buffer_++;
				if(debug_)
				{
					cout << Scheduler::instance().clock() << " Non favor packet dropped, seqno is " << elt.seqno << ", fid is " << elt.id << endl;
				}
				return false;
			}
			else
			{
				*(buffer_reordering_+full) = elt;
				buffer_reordering_filling_++;
				if(debug_)
				{
					cout << Scheduler::instance().clock() << Scheduler::instance().clock() << " HARQR: packet put in buffer " << elt.seqno << endl;
				}
				return true;
			}
		}
	}
	else
	{
		for(i=0;i<buffer_reordering_size_;i++)
		{
			if((buffer_reordering_+i)->valid == false)
			{
				*(buffer_reordering_+i) = elt;
				buffer_reordering_filling_++;
				if(debug_)
				{
					cout << Scheduler::instance().clock() << " HARQR: packet put in buffer " << elt.seqno << endl;
				}
				return true;
			}
		}
		if(debug_)
		{
			cout << " HARQR: buffer out full" << endl;
		}
		if(trace_reordering_)
		{
			print_trace_reordering(elt.p,'d',0,1);
		}
		drop(elt.p);
		dropped_reordering_buffer_++;
		return false;
	}
}

bool HARQR::deque(pkt_recv* pkt, int expected, int fid)
{
	int i;
	
	if(debug_)
	{
		cout << Scheduler::instance().clock() << " HARQR: searching pkt with seqno " << expected << endl;
	}
	for(i=0;i<buffer_reordering_size_;i++)
	{
		if((buffer_reordering_+i)->valid)
		{
			if(((buffer_reordering_+i)->seqno == expected || expected == -1) && (buffer_reordering_+i)->id == fid)
			{
				*pkt = *(buffer_reordering_+i);
				(buffer_reordering_+i)->valid = false;
				buffer_reordering_filling_--;
				if(debug_)
				{
					cout << Scheduler::instance().clock() << " HARQR: \t->found!" << endl;
				}
				if(expected != -1){
					return true;
				}
				if(trace_reordering_)
				{
					print_trace_reordering(pkt->p,'d',isFavor(pkt->id,1),2);
					Packet::free(pkt->p);
					dropped_reordering_buffer_++;
				}
			}
		}
	}
	if(debug_)
	{
		cout << Scheduler::instance().clock() << " HARQR: \t-> not found!" << endl;
	}
	return false;
}

void HARQR::print_trace_reconstruction(Packet* p, int type, char status,int decoded, int retrans, int size){
	hdr_cmn *th = hdr_cmn::access(p);
	hdr_tcp *tcph = hdr_tcp::access(p);
	hdr_ip *iph = hdr_ip::access(p);
	sprintf(pt_reconstruction_->buffer(), "%c "TIME_FORMAT" %s %d %d %d %d %d",
		status,
		pt_reconstruction_->round(Scheduler::instance().clock()),
		packet_info.name(type),
		iph->flowid(),
		tcph->seqno_,
		decoded,
		size == 0 ? th->size() : size,
		retrans);
	pt_reconstruction_->dump();
}

void HARQR::print_trace_reordering(Packet* p, char status,int favor, int reason){
	if(p != 0){
		hdr_cmn *th = hdr_cmn::access(p);
		hdr_tcp *tcph = hdr_tcp::access(p);
		hdr_ip *iph = hdr_ip::access(p);
		int type = th->ptype_;
		sprintf(pt_reordering_->buffer(), "%c "TIME_FORMAT" %s %d %d %d %d %d %d",
			status,
			pt_reordering_->round(Scheduler::instance().clock()),
			packet_info.name(type),
			iph->flowid(),
			tcph->seqno_,
			th->size(),
			favor,
			reason,
			buffer_reordering_filling_);
		pt_reordering_->dump();
	}else{
		sprintf(pt_reordering_->buffer(), "%c "TIME_FORMAT" %d %d %d %d %d %d %d",
			status,
			pt_reordering_->round(Scheduler::instance().clock()),
			0,
			0,
			0,
			0,
			0,
			0,
			buffer_reordering_filling_);
		pt_reordering_->dump();
	}
}

double HARQR::getError(int nb_trans, double IM, int offset)
{
	cout << "Parent method of getError(), should not be called" << endl;
	return 0;
}

int HARQR::computeBitsToSend(int nb_trans)
{	
	cout << "Parent method of computeBitsToSend(), should not be called" << endl;
	return 0;
}

void HARQR::results()
{
	int i;
	int total_forward = 0;
	for(i=0;i<=RETRY_LIMIT;i++)
	{
		total_forward+=forwarded_[i];
	}
	
	cout << "Packets dropped by reconstruction buffer: " << dropped_ << " (" << dropped_*100.0/total_forward << "%)" << endl;
	for(i=0;i<=RETRY_LIMIT;i++)
	{
		cout << "Packets forwarded with " << i << " retransmission(s): " << forwarded_[i] << " (" << forwarded_[i]*100.0/total_forward << "%)" << endl;
	}
	
	cout << "Packets dropped by reordering buffer: " << dropped_reordering_buffer_ << endl;
}
