#--------------------------------------------------------------------------------
# Author: Bastien Tauran
# Updated: 4-11-2016
#--------------------------------------------------------------------------------

######################################
#          Global functions          #
######################################

SimpleLink instproc link-harq {} {
	global ns n1 n2
	$self instvar link_ queue_ drophead_
	$self instvar tHARQ_ rHARQ_

	set tHARQ_ [new Agent/HARQT]
	set rHARQ_ [new Agent/HARQR/Static]
	$rHARQ_ set-sender $tHARQ_
	$tHARQ_ target [$queue_ target] 
	$queue_ target $tHARQ_
	$tHARQ_ drop-target $drophead_
    $rHARQ_ target [$link_ target]
    $link_ target $rHARQ_
    $link_ drop-target $drophead_
	
	set qu [[$ns link $n1 $n2] queue]
	$tHARQ_ set-queue $qu
	$qu set-harq $tHARQ_ $rHARQ_
    
    return $rHARQ_
}

Simulator instproc link-harq {from to} {
	set link [$self link $from $to]
	set ret [$link link-harq]
	return $ret
}

proc change_delay {new_value1 new_value2 i} {
	global ns n_s n1 n2 rHARQ_ EsN0

	set new_value1_ms [expr $new_value1*1000]
	append new_value1_ms "ms"
	$ns delay $n_s($i) $n1 $new_value1_ms duplex
	
	if {$i == 0} {
		set new_value2_ms [expr $new_value2*1000]
		append new_value2_ms "ms"
		$ns delay $n1 $n2 $new_value2_ms duplex
		if {$EsN0 != 0} {
			$rHARQ_ set-delay $new_value2
		}
	}
}

#######################################
#          Getting arguments          #
#######################################

set USAGE "ns iridium-harq-multi.tcl <nodes> <scheduling> <Es/N0> <offset>"
if {$argc != 4} {puts $USAGE; exit 0}

set nodes		[lindex $argv 0];
set scheduling	[lindex $argv 1];
set EsN0		[expr {int([lindex $argv 2])}];
set offset		[lindex $argv 3];

set duration 100

########################################
#          Creating simulator          #
########################################

Agent/HARQT 				set debug_ 					0
Agent/HARQR 				set debug_ 					0
Queue/HARQQueue				set debug_ 					0
Queue/HARQQueue/RR			set debug_ 					0
Queue/HARQQueue/MLWDF		set debug_ 					0
Agent/HARQR 				set reordering_ 			0
Agent/HARQR 				set trace_reconstruction_	1
Agent/HARQR 				set trace_reordering_		1
Queue 						set limit_ 					100

set ns [new Simulator]
$ns color 1 Red

###################################################
#          Configuring global parameters          #
###################################################

#Open the NAM trace file
set nf [open nam/harq.nam w]
$ns namtrace-all $nf

#Open the Trace file
append name_tr "traces/iridium-harq-multi-UDP-" $nodes "-" $scheduling "-" $EsN0  "-" $offset ".tr"
set tf [open $name_tr w]
$ns trace-all $tf

#Define a 'finish' procedure
proc finish {} {
	puts "Simulation finished"

	global ns nf tf
	$ns flush-trace
	#Close the NAM trace file
	close $nf
	#Close the Trace file
	close $tf
	#Execute NAM on the trace file
	#exec nam nam/harq.nam &
	exit 0
}

##################################
#          Set up nodes          #
##################################

set nb_trans $nodes

for {set i 0} {$i < $nb_trans} {incr i} {
    set n_s($i) [$ns node]
}

set n1 [$ns node]
set n2 [$ns node]

for {set i 0} {$i < $nb_trans} {incr i} {
    set n_r($i) [$ns node]
}

for {set i 0} {$i < $nb_trans} {incr i} {
    $ns duplex-link $n_s($i) $n1 50Mb 100ms DropTail
}

if {$scheduling == "MLWDF"} {
	$ns simplex-link $n1 $n2 50Mb 10ms HARQQueue/MLWDF
} elseif {$scheduling == "PF"} {
	$ns simplex-link $n1 $n2 50Mb 10ms HARQQueue/PF
} elseif {$scheduling == "PFDrop"} {
	$ns simplex-link $n1 $n2 50Mb 10ms HARQQueue/PFDrop
} elseif {$scheduling == "RR"} {
	$ns simplex-link $n1 $n2 50Mb 10ms HARQQueue/RR
} elseif {$scheduling == "FIFO_HIGH"} {
	$ns simplex-link $n1 $n2 50Mb 10ms HARQQueue/FIFO
} elseif {$scheduling == "FIFO_LOW"} {
	$ns simplex-link $n1 $n2 50Mb 10ms HARQQueue/FIFO
} else {
	puts "Wrong scheduler"
	exit 0
}

$ns simplex-link $n2 $n1 55Mb 10ms DropTail

for {set i 0} {$i < $nb_trans} {incr i} {
	$ns duplex-link $n2 $n_r($i) 50Mb 1ms DropTail
	source iridium-delay-changes.tcl
}
  
#Open the queue trace file
append name_queue_tr "traces/iridium-harq-multi-UDP-" $nodes "-" $scheduling "-" $EsN0  "-" $offset "-queue.tr"
set tqueue [open $name_queue_tr w]
set qu [[$ns link $n1 $n2] queue]
$qu trace-queue $tqueue

set sqr [expr { sqrt($nb_trans) }]
set qsize [expr 10000000/$sqr/1680]

if {$scheduling == "FIFO_HIGH"} {
	$qu set-queue-size $qsize
} else {
	$qu set-queue-size [expr $qsize/$nb_trans]
}
puts $qsize

$qu set-PFDrop-timeout 0.2

#####################################
#          Set up  agents           #
#####################################

for {set i 0} {$i < $nb_trans} {incr i} {
    #Setup a TCP connection
	set udp($i) [new Agent/UDP]
	$ns attach-agent $n_s($i) $udp($i)
	set null($i) [new Agent/Null]
	$ns attach-agent $n_r($i) $null($i)
	$ns connect $udp($i) $null($i)
	$udp($i) set fid_ $i

	#Setup a FTP over TCP connection
	set VoIP 0
	if {$VoIP == 1} {
		set traffic($i) [new Application/Traffic/Pareto]
		$traffic($i) attach-agent $udp($i)
		$traffic($i) set type_ pareto
		$traffic($i) set packetsize_ 210
		$traffic($i) set burst_time_ 500ms
		$traffic($i) set idle_time_ 50ms
	} else {
		set traffic($i) [new Application/Traffic/CBR]
		$traffic($i) attach-agent $udp($i)
		$traffic($i) set type_ CBR
		$traffic($i) set random_ false
		$traffic($i) set packet_size_ 210
		$traffic($i) set rate_ 1M
	}

	#Schedule events for the FTP agent
	$ns at [expr $offset + 0.1*$i+0.5] "$traffic($i) start"
	$ns at [expr $offset + 100.0] "$traffic($i) stop"
	$ns at [expr $offset + 100.5] "$ns detach-agent $n_s($i) $udp($i) ; $ns detach-agent $n_r($i) $null($i)"
}

#Set HARQ
if {$EsN0 != 0} {
	puts "HARQ enabled"
	set rHARQ_ [$ns link-harq $n1 $n2]
	$rHARQ_ set-delay 0.01
	$rHARQ_ set-EsN0 $EsN0
	$rHARQ_ set-buffer-reordering-size 125
	$rHARQ_ set-buffer-fid-size 1250

	#Open the reconstruction trace file
	append name_rec_tr "traces/iridium-harq-multi-UDP-" $nodes "-" $scheduling "-" $EsN0  "-" $offset "-reconstruction.tr"
	set trf [open $name_rec_tr w]
	$rHARQ_ trace-reconstruction $trf
}

###################################################
#          Last configuration parameters          #
###################################################

#Finish simulation
if {$EsN0 != 0} {
	$ns at [expr $offset + 100.5] "$rHARQ_ results"
	$ns at [expr $offset + 100.5] "$qu results"
}
$ns at [expr $offset + 101.0] "finish"

puts "start simulation"

$ns run
