#--------------------------------------------------------------------------------
# Author: Bastien Tauran
# Updated: 4-11-2016
#--------------------------------------------------------------------------------

######################################
#          Global functions          #
######################################

SimpleLink instproc link-harq {} {
	global ns n1 n2
	$self instvar link_ queue_ drophead_
	$self instvar tHARQ_ rHARQ_

	set tHARQ_ [new Agent/HARQT]
	set rHARQ_ [new Agent/HARQR/Static]
	$rHARQ_ set-sender $tHARQ_
	$tHARQ_ target [$queue_ target] 
	$queue_ target $tHARQ_
	$tHARQ_ drop-target $drophead_
    $rHARQ_ target [$link_ target]
    $link_ target $rHARQ_
    $link_ drop-target $drophead_
	
	set qu [[$ns link $n1 $n2] queue]
	$tHARQ_ set-queue $qu
	$qu set-harq $tHARQ_ $rHARQ_
    
    return $rHARQ_
}

Simulator instproc link-harq {from to} {
	set link [$self link $from $to]
	set ret [$link link-harq]
	return $ret
}

proc plotWindow {tcpSource outfile i} {
	global ns

	set now [$ns now]
	set cwnd [$tcpSource set cwnd_]

	puts $outfile "$now $i $cwnd"

	$ns at [expr $now+0.1] "plotWindow $tcpSource $outfile $i"
}

proc plotRTO {tcpSource outfile i} {
	global ns

	set now [$ns now]
	set rtt [$tcpSource set srtt_]
	set rttvar [$tcpSource set rttvar_]

	puts $outfile "$now $i [expr $rtt/8] [expr $rttvar/4] [expr $rtt/8+$rttvar]"

	$ns at [expr $now+0.1] "plotRTO $tcpSource $outfile $i"
}

proc change_delay {new_value1 new_value2 i} {
	global ns n_s n1 n2 rHARQ_ EsN0

	set new_value1_ms [expr $new_value1*1000]
	append new_value1_ms "ms"
	$ns delay $n_s($i) $n1 $new_value1_ms duplex
	
	if {$i == 0} {
		set new_value2_ms [expr $new_value2*1000]
		append new_value2_ms "ms"
		$ns delay $n1 $n2 $new_value2_ms duplex
		if {$EsN0 != 0} {
			$rHARQ_ set-delay $new_value2
		}
	}
}

#######################################
#          Getting arguments          #
#######################################

set USAGE "ns iridium-harq-multi.tcl <protocol> <nodes> <scheduling> <Es/N0> <offset> <maxRTO> <timeout>"
if {$argc != 7} {puts $USAGE; exit 0}

set protocol	[lindex $argv 0];
set nodes		[lindex $argv 1];
set scheduling	[lindex $argv 2];
set EsN0		[expr {int([lindex $argv 3])}];
set offset		[lindex $argv 4];
set maxRTO		[lindex $argv 5];
set timeout		[lindex $argv 6];

# Duration of simulation
set duration 100

########################################
#          Creating simulator          #
########################################

Agent/HARQT 				set debug_ 					0
Agent/HARQR 				set debug_ 					0
Queue/HARQQueue				set debug_ 					0
Queue/HARQQueue/RR			set debug_ 					0
Queue/HARQQueue/MLWDF		set debug_ 					0
Agent/HARQR 				set reordering_ 			0
Agent/HARQR 				set trace_reconstruction_	1
Agent/HARQR 				set trace_reordering_		1
Agent/TCP 					set window_ 				50000
Agent/TCPSink 				set generateDSacks_ 		false
Queue 						set limit_ 					100

Agent/TCP/Linux				set relentless_				0

if {$maxRTO != 0} {
	Agent/TCP/Linux 			set maxrto_ 				$maxRTO
}

set ns [new Simulator]
$ns color 1 Red

###################################################
#          Configuring global parameters          #
###################################################

#Open the NAM trace file
set nf [open nam/harq.nam w]
$ns namtrace-all $nf

#Open the Trace file
append name_tr "traces/iridium-harq-multi-" $protocol "-" $nodes "-" $scheduling "-" $EsN0 "-" $offset "-" $maxRTO "-" $timeout ".tr"
set tf [open $name_tr w]
$ns trace-all $tf

#Open the cwnd file
append name_cwnd "traces/iridium-harq-multi-" $protocol "-" $nodes "-" $scheduling "-" $EsN0 "-" $offset "-" $maxRTO "-" $timeout ".cwnd"
set cwnd_file  [open $name_cwnd w]

#Open the rto file
append name_rto "traces/iridium-harq-multi-" $protocol "-" $nodes "-" $scheduling "-" $EsN0 "-" $offset "-" $maxRTO "-" $timeout ".rto"
set rto_file  [open $name_rto w]

#Define a 'finish' procedure
proc finish {} {
	puts "Simulation finished"

	global ns nf tf cwnd_file rto_file
	$ns flush-trace
	#Close the NAM trace file
	close $nf
	#Close the Trace file
	close $tf
	#Close the cwnd file
	close $cwnd_file
	#Close the rto file
	close $rto_file
	#Execute NAM on the trace file
	#exec nam nam/harq.nam &
	exit 0
}

##################################
#          Set up nodes          #
##################################

set nb_trans $nodes

for {set i 0} {$i < $nb_trans} {incr i} {
    set n_s($i) [$ns node]
}

set n1 [$ns node]
set n2 [$ns node]

for {set i 0} {$i < $nb_trans} {incr i} {
    set n_r($i) [$ns node]
}

for {set i 0} {$i < $nb_trans} {incr i} {
    $ns duplex-link $n_s($i) $n1 50Mb 100ms DropTail
    source iridium-delay-changes.tcl
}

if {$scheduling == "MLWDF"} {
	$ns simplex-link $n1 $n2 50Mb 10ms HARQQueue/MLWDF
} elseif {$scheduling == "PF"} {
	$ns simplex-link $n1 $n2 50Mb 10ms HARQQueue/PF
} elseif {$scheduling == "PFDrop"} {
	$ns simplex-link $n1 $n2 50Mb 10ms HARQQueue/PFDrop
} elseif {$scheduling == "RR"} {
	$ns simplex-link $n1 $n2 50Mb 10ms HARQQueue/RR
} elseif {$scheduling == "FIFO_HIGH"} {
	$ns simplex-link $n1 $n2 50Mb 10ms HARQQueue/FIFO
} elseif {$scheduling == "FIFO_LOW"} {
	$ns simplex-link $n1 $n2 50Mb 10ms HARQQueue/FIFO
} else {
	puts "Wrong scheduler"
	exit 0
}

$ns simplex-link $n2 $n1 55Mb 10ms DropTail

for {set i 0} {$i < $nb_trans} {incr i} {
	$ns duplex-link $n2 $n_r($i) 50Mb 1ms DropTail
}
  
#Open the queue trace file
append name_queue_tr "traces/iridium-harq-multi-" $protocol "-" $nodes "-" $scheduling "-" $EsN0 "-" $offset "-" $maxRTO "-" $timeout "-queue.tr"
set tqueue [open $name_queue_tr w]
set qu [[$ns link $n1 $n2] queue]
$qu trace-queue $tqueue

set sqr [expr { sqrt($nb_trans) }]
set qsize [expr 10000000/$sqr/1680]

if {$scheduling == "FIFO_HIGH"} {
	$qu set-queue-size $qsize
} else {
	$qu set-queue-size [expr $qsize/$nb_trans]
}
puts $qsize

$qu set-PFDrop-timeout $timeout

#####################################
#          Set up  agents           #
#####################################

for {set i 0} {$i < $nb_trans} {incr i} {
    #Setup a TCP connection
	set tcp($i) [new Agent/TCP/Linux]
	$tcp($i) select_ca $protocol
	$ns attach-agent $n_s($i) $tcp($i)
	set sink($i) [new Agent/TCPSink/Sack1]
	$ns attach-agent $n_r($i) $sink($i)
	$ns connect $tcp($i) $sink($i)
	$tcp($i) set fid_ $i

	#Setup a FTP over TCP connection
	set ftp($i) [new Application/FTP]
	$ftp($i) attach-agent $tcp($i)
	$ftp($i) set type_ FTP

	#Schedule events for the FTP agent
	$ns at [expr $offset + 0.5*$i+0.5] "$ftp($i) start"
	$ns at [expr $offset + 600.0] "$ftp($i) stop"
	$ns at [expr $offset + 600.5] "$ns detach-agent $n_s($i) $tcp($i) ; $ns detach-agent $n_r($i) $sink($i)"
	
	$ns at 0.0 "plotWindow $tcp($i) $cwnd_file $i"
	$ns at 0.0 "plotRTO $tcp($i) $rto_file $i"
}

#Set HARQ
if {$EsN0 != 0} {
	puts "HARQ enabled"
	set rHARQ_ [$ns link-harq $n1 $n2]
	$rHARQ_ set-delay 0.01
	$rHARQ_ set-EsN0 $EsN0
	$rHARQ_ set-buffer-reordering-size 125
	$rHARQ_ set-buffer-fid-size 1250

	#Open the reconstruction trace file
	append name_rec_tr "traces/iridium-harq-multi-" $protocol "-" $nodes "-" $scheduling "-" $EsN0 "-" $offset "-" $maxRTO "-" $timeout "-reconstruction.tr"
	set trf [open $name_rec_tr w]
	$rHARQ_ trace-reconstruction $trf
	
	#Open the reordering trace file
	append name_reo_tr "traces/iridium-harq-multi-" $protocol "-" $nodes "-" $scheduling "-" $EsN0 "-" $offset "-" $maxRTO "-" $timeout "-reordering.tr"
	set tref [open $name_reo_tr w]
	$rHARQ_ trace-reordering $tref
}

###################################################
#          Last configuration parameters          #
###################################################

#Finish simulation
if {$EsN0 != 0} {
	$ns at [expr $offset + 600.5] "$rHARQ_ results"
	$ns at [expr $offset + 600.5] "$qu results"
}
$ns at [expr $offset + 601.0] "finish"

puts "start simulation"

$ns run
